from typing import Any, Callable

Predicate = Callable[[Any], bool]


class Validator(object):
    """
    Validates given value using provided predicates.

    .. automethod:: __call__
    """

    def __init__(self, *predicates: Predicate):
        """
        Instantiates the validator.

        :param predicates: predefined predicates
        :type predicates: Callable[[Any], bool]
        """
        self.predicates = list(predicates)

    def add(self, predicate: Predicate) -> "Validator":
        """
        Adds new predicate to the list.

        :param predicate: predicate function
        :return: self
        """
        self.predicates.append(predicate)
        return self

    def __call__(self, value: Any) -> bool:
        """
        Provides given value to each predicate.

        :param value: value to validate
        :return: True if all checks are passed, False otherwise
        """
        for p in self.predicates:
            if not p(value):
                return False
        return True



from decimal import Decimal

from django.http import HttpRequest, JsonResponse
from qval import qval, validate

from app.validators import *


def division_view(request: HttpRequest):
    """
    GET /api/divide?
    param a : int
    param b : int, nonzero

    Example: GET /api/divide?a=10&b=2&token=abcdefghijkl -> 200, {"answer": 5}
    """
    # Parameter validation occurs in the context manager.
    # If validation fails or user code throws an error, context manager
    # will raise InvalidQueryParamException or APIException respectively.
    # In Django, these exception will be processed and result
    # in error codes (400 and 500) on the client side.
    params = (
        validate(request, a=int, b=int)
        # `b` must be anything but zero
        .nonzero("b")
    )
    with params as p:
        return JsonResponse({"answer": p.a // p.b})


@qval({"a": float, "b": float})
def pow_view(request, params):
    """
    GET /api/pow?
    param a : float
    param b : float

    Example: GET /api/pow?a=2&b=3 -> 200, {"answer": 8.0}
    Example: GET /api/pow?a=2&b=3000000000000 -> 500,
    {
        "error": "An error occurred while processing you request. Please contact the website administrator."
    }
   """
    # Here we don't catch the OverflowError if `b` is too big.
    # This will result in 500 error on the client side.
    return JsonResponse({"answer": params.a ** params.b})


@qval(purchase_factories, purchase_validators)
def purchase_view(request, params):
    """
    GET /api/purchase?
    param item_id : int, positive
    param price   : float, greater than zero
    param token   : string, length == 12

    Example: GET /api/purchase?item_id=1&price=5.8&token=abcdefghijkl
             -> {"success": "Item '1' has been purchased. Check: 5.92$."
    """
    tax = 0.02
    cost = params.price * Decimal(1 + tax)
    return JsonResponse(
        {
            "success": f"Item '{params.item_id}' has been purchased. Check: {round(cost, 2)}$."
        }
    )



import os
from io import StringIO
import pytest

from qval.utils import get_request_params, log, load_symbol
from tests.crossframework import builder

symbol = lambda x, y: x + y
QVAL_LOGGERS = [lambda name: lambda v: f"[{name}]: {v}"]


def test_get_request_params_rejects_invalid_objects():
    objects = [object(), {}, [], ()]

    for obj in objects:
        with pytest.raises(AttributeError):
            get_request_params(obj)


def test_get_request_params_accepts_supported_requests():
    for request in builder.iterbuild({"test": "request"}):
        assert isinstance(get_request_params(request), dict)


def test_load_symbol():
    func = load_symbol("tests.test_utils.symbol")
    assert func(30, 12) == 42

    func = load_symbol(symbol)
    assert func is symbol
    assert func(30, 12) == 42


@pytest.fixture()
def envvar():
    os.environ["DJANGO_SETTINGS_MODULE"] = "tests.test_utils"
    yield None
    del os.environ["DJANGO_SETTINGS_MODULE"]


def test_log_detect_loggers(envvar):
    logger = log.detect_loggers(silent=True)

    assert len(logger.factories) == 1
    assert callable(logger.factories[0])

    msg = logger.factories[0]("test")("message")
    assert msg == f"[test]: message"


def test_log_add_loggers():
    buf = StringIO()
    logger = log.detect_loggers()

    logger.add_logger(lambda name: lambda v: buf.write(f"[{name}]: {v}\n"))
    logger.dump("test", "info", "message")

    assert buf.getvalue() == "[test]: message\n"


def test_logging_switch():
    buf = StringIO()
    logger = log.detect_loggers()

    logger.add_logger(lambda name: lambda v: buf.write(f"[{name}]: {v}\n"))
    logger.dump("test", "info", "message")
    assert logger.is_enabled
    assert buf.getvalue() == "[test]: message\n"

    logger.disable()
    logger.dump("test", "info", "message")
    assert not logger.is_enabled
    assert buf.getvalue() == "[test]: message\n"



import json
from decimal import Decimal

from falcon import Request, Response, API, HTTP_200
from wsgiref import simple_server
from qval import qval, validate, Validator
from qval.framework_integration import setup_falcon_error_handlers

app = API()

# Setup exception handlers
setup_falcon_error_handlers(app)


class DivisionResource(object):
    def on_get(self, req: Request, resp: Response):
        """
        GET /api/divide?
        param a : int
        param b : int, nonzero

        Example: GET /api/divide?a=10&b=2&token=abcdefghijkl -> 200, {"answer": 5}
        """
        # Parameter validation occurs in the context manager.
        # If validation fails or user code throws an error, context manager
        # will raise InvalidQueryParamException or APIException respectively.
        # In Django, these exception will be processed and result
        # in error codes (400 and 500) on the client side.
        params = (
            validate(req, a=int, b=int)
            # `b` must be anything but zero
            .nonzero("b")
        )
        with params as p:
            resp.status = HTTP_200
            resp.body = json.dumps({"answer": p.a // p.b})


class ExponentiationResource(object):
    @qval({"a": float, "b": float})
    def on_get(self, req: Request, resp: Response, params):
        """
        GET /api/pow?
        param a : float
        param b : float

        Example: GET /api/pow?a=2&b=3 -> 200, {"answer": 8.0}
        Example: GET /api/pow?a=2&b=3000000000000 -> 500,
        {
            "error": "An error occurred while processing you request. Please contact the website administrator."
        }
        """
        # Here we don't catch the OverflowError if `b` is too big.
        # This will result in 500 error on the client side.
        resp.status = HTTP_200
        resp.body = json.dumps({"answer": params.a ** params.b})


class PurchaseResource(object):
    purchase_factories = {"price": Decimal, "item_id": int, "token": None}
    purchase_validators = {
        "price": Validator(lambda x: x > 0),
        "token": Validator(lambda x: len(x) == 12),
        "item_id": Validator(lambda x: x >= 0),
    }

    @qval(purchase_factories, purchase_validators)
    def on_get(self, req: Request, resp: Response, params):
        """
        GET /api/purchase?
        param item_id : int, positive
        param price   : float, greater than zero
        param token   : string, length == 12

        Example: GET /api/purchase?item_id=1&price=5.8&token=abcdefghijkl
                 -> {"success": "Item '1' has been purchased. Check: 5.92$."
        """
        tax = 0.02
        cost = params.price * Decimal(1 + tax)
        resp.status = HTTP_200
        resp.body = json.dumps(
            {
                "success": f"Item '{params.item_id}' has been purchased. Check: {round(cost, 2)}$."
            }
        )


division = DivisionResource()
exponentiation = ExponentiationResource()
purchase = PurchaseResource()

app.add_route("/api/divide", division)
app.add_route("/api/pow", exponentiation)
app.add_route("/api/purchase", purchase)


if __name__ == "__main__":
    httpd = simple_server.make_server("127.0.0.1", 8000, app)
    httpd.serve_forever()



# coding: utf-8
import os
import signal
from subprocess import Popen, PIPE

BASE_DIR = os.path.abspath(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
EX_DIR = os.path.join(BASE_DIR, "examples")


def execute(command: str) -> Popen:
    """
    Executes given command in the `examples` directory.

    :param command: command string
    :return: Popen instance
    """
    cmd = f"cd {EX_DIR} && {command}"
    proc = Popen(cmd, shell=True, preexec_fn=os.setsid)
    proc.exterminate = lambda: os.killpg(os.getpgid(proc.pid), signal.SIGTERM)
    return proc


def start_server(framework: str) -> Popen:
    """
    Starts server for the given framework.
    Supports django, flask and falcon at the moment.

    :param framework: framework name
    :return: opened process
    """
    cmd = f"PYTHONPATH={BASE_DIR}"
    if framework == "django":
        cmd = f"cd django-example && {cmd} python manage.py runserver localhost:8000"
    elif framework == "flask":
        cmd = f"{cmd} FLASK_APP=flask-example.py flask run --port=8000"
    elif framework == "falcon":
        cmd = f"{cmd} python falcon-example.py"
    else:
        raise ValueError(
            f"Unknown framework: {framework}, "
            f"expected `django`, `flask` or `falcon`."
        )
    return execute(cmd)



import functools
from typing import Any, Callable, Dict, Optional, Union
from contextlib import contextmanager, AbstractContextManager, ExitStack

from . import utils
from .validator import Validator
from . import exceptions
from . import framework_integration as fwk


class QueryParamValidator(AbstractContextManager):
    """
    Validates query parameters.

    Examples:
        >>> r = fwk.DummyRequest({"num": "42", "s": "str", "double": "3.14"})
        >>> params = QueryParamValidator(r, dict(num=int, s=None, double=float))
        >>> with params as p:
        ...     print(p.num, p.s, p.double, sep=', ')
        42, str, 3.14


    .. automethod:: __enter__
    .. automethod:: __exit__
    """

    def __init__(
        self,
        request: fwk.Request,
        factories: Dict[str, Optional[type]],
        validators: Dict[str, Validator] = None,
        box_all: bool = True,
    ):
        """
        Instantiates query validator object.

        :param request: fwk.Request instance
        :param factories: mapping of :code:`{param -> factory}`. Providing none as factory is equivalent to :code:`str`
                       or :code:`lambda x: x`, since params are stored as strings.
        :param validators: dictionary of pre-defined validators
        :param box_all: include all params, even if they're not specified in :code:`factories`
        """
        self.request = request
        self._factories = factories
        self._box_all = box_all
        self._query_params = utils.get_request_params(self.request)

        self.result: Dict[str, Any] = {
            k: self.query_params[k]
            # Add all parameters to resulting dictionary if box_all is true.
            # Otherwise keep only specified parameters.
            for k in (self.query_params if self._box_all else self._factories)
        }
        self._params: Dict[str, Validator] = {k: Validator() for k in self.result}
        self._params.update(validators or {})

    @property
    def query_params(self) -> Dict[str, str]:
        """
        Returns dictionary of query parameters.
        """
        return self._query_params

    def add_predicate(self, param: str, predicate: Callable[[Any], bool]):
        """
        Adds new check for provided parameter.

        :param param: name of the request parameter
        :param predicate: predicate function
        :return: None
        """
        if param not in self._params:
            self._params[param] = Validator()
        self._params[param].add(predicate)

    # Alias for add_predicate; returns reference
    def check(
        self, param: str, predicate: Callable[[Any], bool]
    ) -> "QueryParamValidator":
        """
        Adds new check from provided parameter.

        :param param: name of the request parameter
        :param predicate: predicate function
        :return: self
        """
        self.add_predicate(param, predicate)
        return self

    def positive(
        self, param: str, transform: Callable[[Any], Any] = lambda x: x
    ) -> "QueryParamValidator":
        """
        Adds greater than zero comparison check for provided parameter.
        Provided :code:`param` will be tested as [:code:`transform(param) > 0`].

        :param param: name of the request parameter
        :param transform: callable that transforms the parameter, default: :code:`lambda x: x`
        :return: self
        """
        return self.check(param, lambda x: transform(x) >= 0)

    def gt(
        self, param: str, value: Any, transform: Callable[[Any], Any] = lambda x: x
    ) -> "QueryParamValidator":
        """
        Adds greater than comparison check for provided parameter.
        For example, if value = 10, parameter :code:`param` will be tested as [:code:`transform(param) > 10`].

        :param param: name of the request parameter
        :param value: value to compare with
        :param transform: callable that transforms the parameter, default: :code:`lambda x: x`
        :return: self
        """
        return self.check(param, lambda x: transform(x) > value)

    def lt(
        self, param: str, value: Any, transform: Callable[[Any], Any] = lambda x: x
    ) -> "QueryParamValidator":
        """
        Adds less than comparison check for provided parameter.
        For example, if value = 10, parameter :code:`param` will be tested as [:code:`transform(param) < 10`].

        :param param: name of the request parameter
        :param value: value to compare with
        :param transform: callable that transforms the parameter, default: :code:`lambda x: x`
        :return: self
        """
        return self.check(param, lambda x: transform(x) < value)

    def eq(
        self, param: str, value: Any, transform: Callable[[Any], Any] = lambda x: x
    ) -> "QueryParamValidator":
        """
        Adds equality check for provided parameter.
        For example, if value = 10, parameter :code:`param` will be tested as [:code:`transform(param) == 10`].

        :param param: name of the request parameter
        :param value: value to compare with
        :param transform: callable that transforms the parameter, default: :code:`lambda x: x`
        :return: self
        """
        return self.check(param, lambda x: transform(x) == value)

    def nonzero(
        self, param: str, transform: Callable[[Any], Any] = lambda x: x
    ) -> "QueryParamValidator":
        """
        Adds nonzero check for provided parameter.
        For example, if value = 10, parameter :code:`param` will be tested as [:code:`transform(param) != 0`].

        :param param: name of the request parameter
        :param transform: callable that transforms the parameter, default: :code:`lambda x: x`
        :return: self
        """
        return self.check(param, lambda x: transform(x) != 0)

    @contextmanager
    def _cleanup_on_error(self):
        """
        Unwinds the stack in case of an error.
        """
        with ExitStack() as stack:
            stack.push(self)
            yield
            # The validation checks didn't raise an exception
            stack.pop_all()

    def _validate(self):
        """
        Validates the parameters.
        Only KeyError, ValueError and TypeError are handled as expected errors.

        :return: None
        """
        # Firstly cast parameters to required types
        for param, cast in self._factories.items():
            try:
                # If cast is None, just leave parameter a string
                cast = cast or (lambda x: x)
                value = cast(self.query_params[param])
                self.result[param] = value
            # Missing a required parameter
            except KeyError:
                raise exceptions.InvalidQueryParamException(
                    {"error": f"Missing required parameter `{param}`."},
                    status=fwk.HTTP_400_BAD_REQUEST,
                )
            # Invalid cast
            except (ValueError, TypeError):
                expected = "."
                # Expose only built-in types
                if cast in (int, float):
                    expected = f": expected {cast.__name__}."
                raise exceptions.InvalidQueryParamException(
                    {"error": f"Invalid type of the `{param}` parameter{expected}"},
                    status=fwk.HTTP_400_BAD_REQUEST,
                )

        # Run validations on the each parameter
        for param, value in self.result.items():
            validator = self._params[param]
            if not validator(value):
                raise exceptions.InvalidQueryParamException(
                    {"error": f"Invalid `{param}` value: {self.result[param]}."},
                    status=fwk.HTTP_400_BAD_REQUEST,
                )

    def __enter__(self) -> "utils.FrozenBox":
        """
        Runs validation on provided request. See __exit__() for additional info.

        :return: box of validated values.
        """
        # This context manager will unwind stack in case of an error.
        # The __exit__() method will be called with values of the exception raised inside _validate().
        # This allows us handle exceptions both inside _validate() and inside of the context.
        with self._cleanup_on_error():
            self._validate()
        return utils.FrozenBox(self.result)

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        If occurred exception is not an InvalidUriParamException, the exception will be re-raised as APIException,
        which will result in 500 error on the client side.

        :param exc_type: exception type
        :param exc_val: exception instance
        :param exc_tb: exception traceback
        :return: None
        """
        # Report unexpected exceptions
        if exc_type not in (exceptions.InvalidQueryParamException, None):
            body = getattr(self.request, "body", {})
            text = (
                f"An error occurred during the validation or inside of the context: exc `{exc_type}` ({exc_val}).\n"
                f"| Parameters: {self.query_params}\n"
                f"| Body      : {body}\n"
                f"| Exception:\n"
            )
            utils.log.error(
                __name__,
                text,
                extra={
                    "stack": True,
                    "traceback": exc_tb,
                    "request_body": body,
                    "parameters": self.query_params,
                },
                exc_info=(exc_type, exc_val, exc_tb),
            )
            raise exceptions.APIException(
                detail="An error occurred while processing you request. "
                "Please contact the website administrator."
            ) from exc_val


def validate(
    request: Union[fwk.Request, Dict[str, str]],
    validators: Dict[str, Validator] = None,
    box_all: bool = True,
    **factories,
) -> QueryParamValidator:
    """
    Shortcut for QueryParamValidator.

    Examples:
        >>> r = {"num": "42", "s": "str", "double": "3.14"}
        >>> with validate(r, num=int, s=None, double=float) as p:
        ...     print(p.num + p.double, p.s)
        45.14 str

        >>> r = {"price": "43.5$", "n_items": "1"}
        >>> currency2f = lambda x: float(x[:-1])
        >>> params = validate(r, price=currency2f, n_items=int
        ...     ).positive("n_items")  # n_items must be greater than 0
        >>> with params as p:
        ...     print(p.price, p.n_items)
        43.5 1

    :param request: request instance
    :param validators: dictionary of predefined validators
    :param box_all: include all params that no specified in factories in the param box
    :param factories: factories that create python object from string parameters
    :return: QueryParamValidator instance
    """
    # Wrap dictionary with request-like object
    request = utils.make_request(request)
    return QueryParamValidator(request, factories, validators, box_all)


def qval(
    factories: Dict[str, Optional[Callable[[str], Any]]],
    validators: Dict[str, Validator] = None,
    box_all: bool = True,
    request_: fwk.Request = None,
):
    """
    A decorator that validates query parameters.
    The wrapped function must accept request as first parameter
    (or second if it's a method) and `params` as last.

    :param factories: mapping (parameter, callable [str -> Any])
    :param validators: mapping (parameter, validator)
    :param box_all: include all params that no specified in fields in the param box
    :param request_: optional request object that can be provided to validator
    :return: wrapped function
    """
    # Check if decorator is used improperly
    if callable(factories):
        raise TypeError("qval() missing 1 required positional argument: 'factories'")

    def outer(f):
        @functools.wraps(f)
        def inner(*args, **kwargs):
            args = list(args)
            # If default request object is provided, simply use it
            if request_ is not None:
                request = utils.make_request(request_)
                args.insert(0, request)
            # Otherwise check arguments
            elif isinstance(args[0], fwk.RequestType):
                # And construct request from dict
                request = args[0] = utils.make_request(args[0])
            elif isinstance(args[1], fwk.RequestType):
                request = args[1] = utils.make_request(args[1])
            else:
                raise ValueError(
                    "The first argument of the view must be a request-like."
                )
            with validate(request, validators, box_all, **factories) as params:
                return f(*args, params, **kwargs)

        return inner

    return outer


def qval_curry(request: fwk.Request):
    """
    Wraps `qval()` decorator to provide given request object on each call.

    :param request: request instance
    :return: wrapped `qval(..., request_=request)`
    """

    @functools.wraps(qval)
    def outer(*args, **kwargs):
        kwargs.setdefault("request_", request)
        return qval(*args, **kwargs)

    return outer



import pytest
from qval.utils import FrozenBox


@pytest.fixture()
def dct():
    return {"string": "string", "number": 0x420000, "double": 3.141_592}


def test_box_is_immutable(dct):
    box = FrozenBox(dct)

    with pytest.raises(TypeError):
        box["string"] = "new string"
    assert box.string == "string"

    with pytest.raises(TypeError):
        box.number = 0
    assert box.number == 0x420000

    with pytest.raises(TypeError):
        box["new_key"] = 10

    with pytest.raises(TypeError):
        box.new_key = 10


def test_box_contains_is_valid(dct):
    box = FrozenBox(dct)

    assert "string" in box
    assert "number" in box
    assert "double" in box
    assert "random" not in box


def test_box_iter_works(dct):
    box = FrozenBox(dct)

    assert dict(iter(box)) == dct


def test_error_on_unknown_keys(dct):
    box = FrozenBox(dct)

    with pytest.raises(KeyError):
        box.random

    with pytest.raises(KeyError):
        box["key"]


def test_repr_is_valid(dct):
    box = FrozenBox(dct)
    assert list(eval(repr(box))) == list(FrozenBox(dct))


def test_str_representation(dct):
    box = FrozenBox(dct)
    assert str(box) == f"FrozenBox<{dct}>"



# coding: utf-8
import time
import pytest

from tests.test_frameworks.common import start_server
from tests.test_frameworks.api_test_functions import *


# Common tests from `api_test_functions` will be ran on the django server
@pytest.mark.plain_django
@pytest.fixture(scope="module")
def server(request):
    print("Starting django server...")
    proc = start_server("django")
    request.addfinalizer(proc.exterminate)
    # Wait while server is setting up
    time.sleep(1.5)
    return proc



import pytest
from decimal import Decimal

from qval import InvalidQueryParamException, qval, qval_curry
from qval.framework_integration import HTTP_400_BAD_REQUEST, Request
from qval.utils import FrozenBox, dummyfy
from qval.validator import Validator

from tests.crossframework import builder


stats = []
factories = {
    "double": float,
    "num": int,
    "price": Decimal,
    "hashme": hash,
    "observable": stats.append,
}
validators = {"price": Validator(lambda x: x > 0)}


@qval(factories, validators)
def simple_view(request: Request, params: FrozenBox):
    return dummyfy(request), params


class ViewClass(object):
    @qval(factories, validators)
    def complex_view(
        self, request: Request, param1: int, param2: int, params: FrozenBox
    ):
        return dummyfy(request), param1, param2, params


def test_params_provided():
    params = {
        "double": "3.14",
        "num": "10",
        "price": "2.79",
        "hashme": "s$cret_t0k3n",
        "observable": "important metric",
    }
    for request in builder.iterbuild(params):
        # Test simple view
        r, box = simple_view(request)
        assert stats[-1] == params["observable"]
        assert set(r.query_params.keys()) == set(box.__dct__.keys())

        # Test complex view
        r, p1, p2, box = ViewClass().complex_view(request, 1, 2)
        assert (p1, p2) == (1, 2)
        assert stats[-1] == params["observable"]
        assert set(r.query_params.keys()) == set(box.__dct__.keys())


def test_params_validated():
    params = {
        "double": "3.14",
        "num": "10",
        "price": "0",
        "hashme": "s$cret_t0k3n",
        "observable": "important metric",
    }
    for request in builder.iterbuild(params):
        with pytest.raises(InvalidQueryParamException) as e:
            simple_view(request)
        assert stats[-1] == params["observable"]
        assert e.value.status_code == HTTP_400_BAD_REQUEST

        with pytest.raises(InvalidQueryParamException) as e:
            ViewClass().complex_view(request, 1, 2)
        assert stats[-1] == params["observable"]
        assert e.value.status_code == HTTP_400_BAD_REQUEST


# Curries qval with a static request (in this case)
# Useful in frameworks with global request classes (e.g. Flask)
def get_curried_qval():
    request = {
        "double": "3.14",
        "num": "10",
        "price": "2.79",
        "hashme": "s$cret_t0k3n",
        "observable": "important metric",
    }
    return qval_curry(request)


curried_qval = get_curried_qval()


@curried_qval(factories, validators)
def view(request, some_param, params):
    return dummyfy(request), some_param, params


def test_curried_qval():
    # Test simple view
    r, test, box = view("test")
    assert test == "test"
    assert stats[-1] == r.query_params["observable"]
    assert set(r.query_params.keys()) == set(box.__dct__.keys())



import logging
from pprint import pformat
from typing import Dict, Any, List, Callable, Union

from qval.framework_integration import load_symbol
from . import framework_integration as fwk


@fwk._make_request
def make_request(request: Union[Dict[str, str], fwk.Request]) -> fwk.RequestType:
    """
    Creates DummyRequest if :code:`request` is a dictionary, and returns the :code:`request` itself otherwise.

    Behavior of this function can be customized with the
    :func:`@_make_request() <qval.framework_integration._make_request>` decorator.
    Provide path to your wrapper using :code:`QVAL_MAKE_REQUEST_WRAPPER` in the settings file
    or set it as an environment variable. The wrapper function must accept :code:`request` as parameter and
    return object that implements request interface.

    For example, the following code adds print to the each function call:
    ::

        # app/utils.py
        def my_wrapper(f):
            @functools.wraps(f)
            def wrapper(request):
                print(f"Received new request: {request}")
                return f(request)
            return wrapper

    Then execute :code:`export QVAL_MAKE_REQUEST_WRAPPER=app.utils.my_wrapper` in your console
    or simply add it to the config file.

    :param request: dict or request instance
    :return: request
    """
    if isinstance(request, dict):
        return fwk.DummyRequest(request)
    return request


def get_request_params(request: fwk.RequestType):
    """
    Returns dictionary of query parameters of the given request.

    :param request: any supported request
    :return: dictionary of parameters
    """
    supported_attrs = ("query_params", "GET", "args", "params")
    for attr in supported_attrs:
        if hasattr(request, attr):
            return getattr(request, attr)
    raise AttributeError(
        "Provided request object has no any of the following attributes: "
        "{}.".format(", ".join(f"`{attr}`" for attr in supported_attrs))
    )


def dummyfy(request: fwk.Request) -> fwk.DummyRequest:
    """
    Constructs :class:`qval.framework_integration.DummyRequest` with params of the given request.

    :param request: any supported request
    :return: :code:`DummyRequest(request.<params>)`
    """
    return fwk.DummyRequest(get_request_params(request))


class FrozenBox(object):
    """
    Frozen dictionary that allows accessing elements by :code:`.`

    Example:
        >>> box = FrozenBox({"num": 10, "s": "string"})
        >>> print(box.num, box.s)
        10 string
        >>> box["num"] = 404
        Traceback (most recent call last):
            ...
        TypeError: 'FrozenBox' object does not support item assignment
        >>> box.num = 404
        Traceback (most recent call last):
            ...
        TypeError: 'FrozenBox' object does not support attribute assignment
        >>> box.num
        10
    """

    def __init__(self, dct: Dict[Any, Any]):
        """
        :param dct: dict to store
        """
        self.__dict__["__dct__"] = dct

    def __getitem__(self, item: str) -> Any:
        """
        [] operator.

        :param item: item key
        :return: value for key `item`
        """
        return self.__dict__["__dct__"][item]

    def __getattr__(self, item: str) -> Any:
        """
        Returns value of the stored `item` or attribute of the object.

        :param item: item key
        :return: value
        """
        return self[item]

    def __setattr__(self, key: str, value: str):
        """
        Raises TypeError.
        """
        raise TypeError(
            f"'{self.__class__.__name__}' object does not support attribute assignment"
        )

    def __contains__(self, item: str) -> bool:
        """
        Determines if item is inside of the dictionary.
        :param item: item to check
        """
        return item in self.__dict__["__dct__"]

    def __iter__(self):
        """
        Returns iterator over :code:`__dct__.items()`

        :return: :code:`iter(__dct__.items())`
        """
        return iter(self.__dict__["__dct__"].items())

    def __repr__(self) -> str:
        """
        Returns evaluable representation of the :class:`FrozenBox` object.
        """
        return f"FrozenBox({self.__dict__['__dct__']})"

    def __str__(self) -> str:
        """
        Returns string representation of the :class:`FrozenBox` object.

        :return: str(box)
        """
        return f"FrozenBox<{self.__dict__['__dct__']}>"


class ExcLogger(object):
    """
    A class used to report critical errors.

    >>> from qval.utils import log
    >>> log
    ExcLogger([getLogger])
    >>> log.is_enabled
    True
    >>> log.disable()
    >>> print(log)
    ExcLogger<[getLogger], enabled = false>

    """

    def __init__(self, logger_factories: List[Callable[[str], Any]]):
        """
        Instantiates the ExcLogger.

        :param logger_factories: list of logger factories
        """
        self.factories = logger_factories
        self._enabled = True

    @property
    def is_enabled(self) -> bool:
        """
        Returns True if logging is enabled.
        """
        return self._enabled

    def disable(self):
        """
        Disables logging.

        :return: None
        """
        self._enabled = False

    def enable(self):
        """
        Enables logging.

        :return: None
        """
        self._enabled = True

    def add_logger(self, log_factory: Callable[[str], Any]):
        """
        Adds new logger factory to list.

        :param log_factory: logger
        :return: None
        """
        self.factories.append(log_factory)

    def clear(self):
        """
        Removes all saved factories.
        :return: None
        """
        self.factories.clear()

    def dump(self, name: str, level: str, *args, **kwargs):
        """
        Instantiates new loggers using configured factories and provides
        :code:`*args` and :code:`**kwargs` to built objects.

        :param name: logger name
        :param level: logging level. If built object has no attribute :code:`level`, it will be treated as callable.
        :param args: logger args
        :param kwargs: logger kwargs
        :return: None
        """
        if not self._enabled:
            return

        for build in self.factories:
            try:
                logger = build(name)
                if hasattr(logger, level):
                    getattr(logger, level)(*args, **kwargs)
                else:
                    logger(*args, **kwargs)
            except TypeError:
                raise
            except:
                pass

    def error(self, name: str, *args, **kwargs):
        """
        Shortcut for :meth:`dump(name, "error", ...) <qval.utils.ExcLogger.dump>`.

        :param name: logger name
        :param args: logger args
        :param kwargs: logger kwargs
        :return: None
        """
        self.dump(name, "error", *args, **kwargs)

    @staticmethod
    def collect_loggers() -> list:
        """
        Looks for configuration and returns list of detected loggers or :func:`logging.getLogger`.

        :return: list of collected loggers
        """
        module = fwk.get_module()
        if hasattr(module, "QVAL_LOGGERS"):
            _loggers = module.QVAL_LOGGERS
            if not isinstance(_loggers, (tuple, list)):
                _loggers = [_loggers]
            loggers = [load_symbol(log) for log in _loggers]
        else:
            loggers = [logging.getLogger]
        return loggers

    @classmethod
    def detect_loggers(cls, silent: bool = False) -> "ExcLogger":
        """
        Looks for configuration and instantiates ExcLogger with detected loggers
        or default :func:`logging.getLogger`.

        :param silent: omit logging test message
        :return: ExcLogger object
        """
        logger = cls(cls.collect_loggers())
        if not silent:
            logger.dump(__file__, "info", "test message")
        return logger

    def __repr__(self) -> str:
        return f"ExcLogger([{', '.join(map(lambda x: x.__name__, self.factories))}])"

    def __str__(self) -> str:
        return (
            f"ExcLogger<[{', '.join(map(lambda x: x.__name__, self.factories))}], "
            f"enabled = {str(self.is_enabled).lower()}>"
        )


log = ExcLogger.detect_loggers()



import requests

base = "http://localhost:8000/api"


def test_divide_parameters_required(server):
    url = f"{base}/divide"
    assert requests.get(url).status_code == 400
    assert requests.get(f"{url}?a=10").status_code == 400
    assert requests.get(f"{url}?b=10").status_code == 400


def test_divide_parameters_types_validated(server):
    url = f"{base}/divide"
    assert requests.get(f"{url}?a=10&b=2.2").status_code == 400
    assert requests.get(f"{url}?a=str&b=string").status_code == 400


def test_divide_parameters_validated(server):
    url = f"{base}/divide"
    r = requests.get(f"{url}?a=10&b=0")
    assert r.status_code == 400
    assert r.json()["error"] == "Invalid `b` value: 0."


def test_divide_success(server):
    url = f"{base}/divide"
    r = requests.get(f"{url}?a=10&b=5")
    assert r.status_code == 200
    assert r.json()["answer"] == 2


def test_pow_parameters_required(server):
    url = f"{base}/pow"
    assert requests.get(url).status_code == 400
    assert requests.get(f"{url}?a=10").status_code == 400
    assert requests.get(f"{url}?b=10").status_code == 400


def test_pow_parameters_validated(server):
    url = f"{base}/pow"
    assert requests.get(f"{url}?a=string&b=str").status_code == 400


def test_pow_overflow_error(server):
    url = f"{base}/pow"
    r = requests.get(f"{url}?a=2.2324&b=30000000")
    assert r.status_code == 500
    assert (
        r.json()["error"]
        == "An error occurred while processing you request. Please contact the website administrator."
    )


def test_pow_success(server):
    url = f"{base}/pow"
    r = requests.get(f"{url}?a=2&b=10")
    assert r.status_code == 200
    assert r.json()["answer"] == 1024


def test_purchase_parameters_required(server):
    url = f"{base}/purchase"
    params = {"item_id=1", "price=4.2", "token=123456789012"}
    for p1 in params:
        for p2 in params - {p1}:
            assert requests.get(f"{url}?{p1}&{p2}").status_code == 400


def test_purchase_parameters_validated(server):
    url = f"{base}/purchase"
    params = [
        ("-10", "123456789012", "13"),
        ("7723", "12345678901", "71"),
        ("3921", "123456789012", "0"),
    ]
    for p0, p1, p2 in params:
        assert (
            requests.get(f"{url}?itemd_id={p0}&token={p1}&price={p2}").status_code
            == 400
        )


def test_purchase_success(server):
    url = f"{base}/purchase"
    r = requests.get(f"{url}?item_id=10&token=123456789012&price=4.04")
    assert r.status_code == 200
    assert r.json()["success"] == "Item '10' has been purchased. Check: 4.12$."



from typing import Callable, Dict, Any, Iterable

from werkzeug.urls import url_encode
from qval.framework_integration import Request as ReqUnion


class RequestBuilder(object):
    def __init__(self):
        self.builders = {}

    def register(self, name: str, builder: Callable[[Dict[str, Any]], ReqUnion]):
        """
        Adds new callable to the request builder.

        :param name: name of the framework
        :param builder: builder callable
        :return: None
        """
        self.builders[name] = builder

    def build(self, name: str, params: Dict[str, Any]) -> ReqUnion:
        """
        Applies builder with name `name` to given params.

        :param name: name of the builder
        :param params: dict of query  params
        :return: built request class
        """
        return self.builders[name](params)

    def iterbuild(self, params: Dict[str, Any]) -> Iterable[ReqUnion]:
        """
        Builds and yield requests using saved builder.

        :param params: query params
        :return: Request classes
        """
        yield from (self.build(b, params) for b in self.builders)


builder = RequestBuilder()

try:
    import os
    import django.http
    from django.conf import settings, empty

    # Disable previously configured settings
    settings._wrapped = empty
    settings.configure()

    def build_django(params: Dict[str, Any]) -> django.http.HttpRequest:
        """
        Builds django.http.HttpRequest.

        :param params: query params
        :return: built request
        """
        r = django.http.HttpRequest()
        r.GET.update(params)
        return r

    builder.register("django", build_django)
except ImportError:
    pass

try:
    import rest_framework.request as drf

    def build_drf(params: Dict[str, Any]) -> drf.Request:
        """
        Builds rest_framework.request.Request.

        :param params: query params
        :return: built request
        """
        return drf.Request(build_django(params))

    builder.register("drf", build_drf)
except ImportError:
    pass


try:
    import flask

    def build_flask(params: Dict[str, Any]) -> flask.app.Request:
        """
        Builds flask.wrappers.Request

        :param params: query parameters
        :return: build request
        """
        query_string = url_encode(params)
        return flask.app.Request({"QUERY_STRING": query_string})

    builder.register("flask", build_flask)
except ImportError:
    pass

try:
    import falcon

    def build_falcon(params: Dict[str, Any]) -> falcon.Request:
        """
        Builds falcon.request.Request
        :param params:
        :return:
        """
        return falcon.Request(
            {
                "QUERY_STRING": url_encode(params),
                # avoid errors while printing
                "wsgi.input": None,
                "wsgi.errors": None,
                "wsgi.url_scheme": "https://",
                "REQUEST_METHOD": "GET",
                "PATH_INFO": "test",
                "SERVER_NAME": "qvaltests",
                "SERVER_PORT": "433",
            }
        )

    builder.register("falcon", build_falcon)
except ImportError:
    pass



from typing import Union

from . import framework_integration


class InvalidQueryParamException(framework_integration.APIException):
    """
    An error thrown when param fails the validation.
    """

    def __init__(self, detail: Union[dict, str], status: int):
        """
        Instantiates the exception.

        :param detail: dict or string with details
        :param status: status code
        """
        super().__init__(detail)
        self.status_code = status


# Avoid circular imports
APIException = framework_integration.APIException
